<?php

declare(strict_types=1);

namespace Drupal\schemadotorg_report\Controller;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\schemadotorg\SchemaDotOrgMappingInterface;
use Drupal\schemadotorg_starterkit\SchemaDotOrgStarterkitManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Returns responses for Schema.org report relationships routes.
 */
class SchemaDotOrgReportRelationshipsController extends SchemaDotOrgReportControllerBase {


  /**
   * An array of background colors.
   *
   * @see https://coolors.co/eddcd2-fff1e6-fde2e4-fad2e1-c5dedd-dbe7e4-f0efeb-d6e2e9-bcd4e6-99c1de
   */
  protected array $colors = [
    '#eddcd2',
    '#fff1e6',
    '#fde2e4',
    '#fad2e1',
    '#c5dedd',
    '#dbe7e4',
    '#f0efeb',
    '#d6e2e9',
    '#bcd4e6',
    '#99c1de',
  ];

  /**
   * An array of hierarchy Schema.org properties.
   */
  protected array $hierarchyProperties = [
    'subOrganization',
    'parentOrganization',
    'subEvent',
    'superEvent',
    'containedInPlace',
    'containsPlace',
    'offeredBy',
    'makesOffer',
    'isPartOf',
    'hasPart',
    'hasVariant',
    'isVariantOf',
    'partOfTrip',
    'subTrip',
    'partOfSeason',
    'partOfSeries',
    'episode',
  ];

  /**
   * An array of relationship Schema.org properties.
   */
  protected array $relationshipProperties = [
    'about',
    'contactPoint',
    'department',
    'employee',
    'healthCondition',
    'member',
    'memberOf',
    'relatedDrug',
    'study',
    'subjectOf',
    'workLocation',
    'worksFor',
  ];

  /**
   * The entity field manager.
   */
  protected EntityFieldManagerInterface $entityFieldManager;

  /**
   * The starter kit manager.
   */
  protected ?SchemaDotOrgStarterkitManagerInterface $starterKitManager = NULL;

  /**
   * A mapping for starter kit type to starter kit names.
   */
  protected array $starterKitTypes;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): static {
    $instance = parent::create($container);
    $instance->entityFieldManager = $container->get('entity_field.manager');

    if ($container->has('schemadotorg_starterkit.manager')) {
      $instance->starterKitManager = $container->get('schemadotorg_starterkit.manager');
    }

    return $instance;
  }

  /**
   * Builds a table containing Schema.org relationships.
   *
   * @return array
   *   A renderable array containing a table containing Schema.org relationships.
   */
  public function overview(): array {
    // Header.
    $header = [];
    $header['label'] = [
      'data' => $this->t('Label'),
      'style' => 'min-width: 200px',
    ];
    $header['id'] = [
      'data' => $this->t('ID'),
      'style' => 'min-width: 100px',
    ];
    $header['description'] = [
      'data' => $this->t('Description'),
      'style' => 'min-width: 400px',
    ];
    if ($this->starterKitManager) {
      $header['starterkit'] = [
        'data' => $this->t('Starter kit'),
        'style' => 'min-width: 100px',
      ];
    }
    $header['type'] = [
      'data' => $this->t('Schema.org type'),
      'style' => 'min-width: 100px',
    ];
    $relationship_types = $this->getRelationshipTypes();
    foreach ($relationship_types as $relationship_type => $relationship_label) {
      $header[$relationship_type] = [
        'data' => $relationship_label['plural'],
        'style' => 'min-width: 100px',
      ];
    }

    // Rows.
    /** @var \Drupal\schemadotorg\SchemaDotOrgMappingStorageInterface $mapping_storage */
    $mapping_storage = $this->entityTypeManager()
      ->getStorage('schemadotorg_mapping');
    /** @var \Drupal\node\Entity\NodeType[] $node_types */
    $node_types = $this->entityTypeManager()
      ->getStorage('node_type')
      ->loadMultiple();

    $rows = [];
    foreach ($node_types as $bundle => $node_type) {
      /** @var \Drupal\schemadotorg\SchemaDotOrgMappingInterface|null $mapping */
      $mapping = $mapping_storage->load("node.$bundle");
      if (!$mapping) {
        continue;
      }

      $row = [];
      $row['label'] = [
        'data' => $node_type->toLink(NULL, 'edit-form')->toRenderable(),
      ];
      $row['id'] = [
        'data' => ['#markup' => $node_type->id()],
      ];
      $row['description'] = [
        'data' => ['#markup' => $node_type->getDescription()],
      ];
      if ($this->starterKitManager) {
        $row['starterkit'] = [
          'data' => [
            '#theme' => 'item_list',
            '#items' => $this->getStarterKits($mapping),
          ],
        ];
      }
      $row['type'] = [
        'data' => [
          '#theme' => 'item_list',
          '#items' => $this->schemaTypeBuilder->buildItemsLinks(
            $mapping->getAllSchemaTypes(),
            ['prefix' => NULL]
          ),
        ],
      ];
      $relationships = $this->getRelationships($mapping);
      foreach ($relationships as $relationship_type => $relationship_items) {
        $row[$relationship_type] = [
          'data' => [
            '#theme' => 'item_list',
            '#items' => $this->schemaTypeBuilder->buildItemsLinks(
              $relationship_items,
              ['prefix' => NULL]
            ),
          ],
        ];
      }
      $rows[] = $row;
    }

    $build = [];
    $build['table'] = [
      '#type' => 'table',
      '#header' => $header,
      '#rows' => $rows,
      '#sticky' => TRUE,
      '#empty' => $this->t('No content types found.'),
      '#attributes' => ['class' => ['schemadotorg-report-table']],
    ];
    $build['#attached']['library'][] = 'schemadotorg_report/schemadotorg_report';
    return $build;
  }

  /**
   * Builds a table containing Schema.org relationship targets.
   *
   * @return array
   *   A renderable array containing a table containing Schema.org relationship targets.
   */
  public function targets(): array {
    // Header.
    $header = [];
    $header['label'] = [
      'data' => $this->t('Label'),
      'style' => 'min-width: 200px',
    ];
    $header['schema_type'] = [
      'data' => $this->t('Schema.org type'),
      'style' => 'min-width: 100px',
    ];
    $header['schema_property'] = [
      'data' => $this->t('Schema.org property'),
      'style' => 'min-width: 100px',
    ];
    $header['relationship_type'] = [
      'data' => $this->t('Relationship type'),
      'style' => 'min-width: 100px',
    ];
    $header['unlimited'] = [
      'data' => $this->t('Unlimited values'),
      'style' => 'min-width: 100px',
    ];
    $header['required'] = [
      'data' => $this->t('Required field'),
      'style' => 'min-width: 100px',
    ];
    $header['targets'] = [
      'data' => $this->t('Target types'),
      'style' => 'min-width: 100px',
    ];
    $header['allowed_values'] = [
      'data' => $this->t('Allowed values'),
      'style' => 'min-width: 100px',
    ];

    $relationship_types = $this->getRelationshipTypes();
    /** @var \Drupal\schemadotorg\SchemaDotOrgMappingStorageInterface $mapping_storage */
    $mapping_storage = $this->entityTypeManager()
      ->getStorage('schemadotorg_mapping');
    /** @var \Drupal\node\Entity\NodeType[] $node_types */
    $node_types = $this->entityTypeManager()
      ->getStorage('node_type')
      ->loadMultiple();

    $index = 0;
    $row_colors = [];
    $rows = [];
    foreach ($node_types as $bundle => $node_type) {
      /** @var \Drupal\schemadotorg\SchemaDotOrgMappingInterface|null $mapping */
      $mapping = $mapping_storage->load("node.$bundle");
      if (!$mapping) {
        continue;
      }

      $schema_properties = $mapping->getAllSchemaProperties();
      $field_definitions = $this->entityFieldManager
        ->getFieldDefinitions('node', $mapping->getTargetBundle());
      foreach ($field_definitions as $field_name => $field_definition) {
        $schema_property = $schema_properties[$field_name] ?? NULL;
        if (!$schema_property) {
          continue;
        }

        $relationship_type = $this->getRelationshipType($schema_property, $field_definition);
        if (!$relationship_type) {
          continue;
        }

        $row = [];
        $row['label'] = [
          'data' => $node_type->toLink(NULL, 'edit-form')->toRenderable(),
        ];
        $row['schema_type'] = [
          'data' => [
            '#theme' => 'item_list',
            '#items' => $this->schemaTypeBuilder->buildItemsLinks(
              $mapping->getAllSchemaTypes(),
              ['prefix' => NULL]
            ),
          ],
        ];
        $row['schema_property'] = [
          'data' => [
            '#type' => 'link',
            '#title' => $schema_property,
            '#url' => $this->schemaTypeBuilder->getItemUrl($schema_property),
          ],
        ];
        $row['relationship_type'] = $relationship_types[$relationship_type]['singular'];
        $row['unlimited'] = $field_definition->getFieldStorageDefinition()->getCardinality() ? $this->t('Yes') : $this->t('No');
        $row['required'] = $field_definition->isRequired() ? $this->t('Yes') : $this->t('No');
        $row['targets'] = [
          'data' => [
          '#theme' => 'item_list',
          '#items' => $this->schemaTypeBuilder->buildItemsLinks(
              $this->getTargets($field_definition),
              ['prefix' => NULL]
            ),
          ],
        ];
        $row['allowed_values'] = [
          'data' => [
            '#theme' => 'item_list',
            '#items' => options_allowed_values($field_definition->getFieldStorageDefinition()) ?: [],
          ],
        ];

        if (!isset($row_colors[$bundle])) {
          $row_colors[$bundle] = $this->colors[$index];
          $index = ($index + 1) % count($this->colors);
        }

        $rows["$bundle-$relationship_type-$schema_property"] = [
          'data' => $row,
          'style' => 'background-color:' . $row_colors[$bundle],
        ];
      }
    }
    ksort($rows);

    $build = [];
    $build['table'] = [
      '#type' => 'table',
      '#header' => $header,
      '#rows' => $rows,
      '#sticky' => TRUE,
      '#empty' => $this->t('No content types found.'),
      '#attributes' => ['class' => ['schemadotorg-report-table']],
    ];
    $build['#attached']['library'][] = 'schemadotorg_report/schemadotorg_report';
    return $build;
  }

  /**
   * Get installed starter kits that sets up the Schema.org mapping type.
   *
   * @param \Drupal\schemadotorg\SchemaDotOrgMappingInterface $mapping
   *   THe Schema.org mapping.
   *
   * @return array
   *   Starter kits that sets up the Schema.org mapping type.
   */
  protected function getStarterKits(SchemaDotOrgMappingInterface $mapping): array {
    if (!$this->starterKitManager) {
      return [];
    }

    // Build starter kit type cache.
    if (!isset($this->starterKitTypes)) {
      $starterkits = $this->starterKitManager->getStarterkits(TRUE);
      $this->starterKitTypes = [];
      foreach ($starterkits as $module_name => $starterkit) {
        $settings = $this->starterKitManager->getStarterkitSettings($module_name);
        foreach (array_keys($settings['types']) as $type) {
          $this->starterKitTypes += [$type => []];
          $this->starterKitTypes[$type][$module_name] = str_replace('Schema.org Blueprints Starter Kit: ', '', $starterkit['name']);
        }
      }
    }

    $entity_type_id = $mapping->getTargetEntityTypeId();
    $bundle = $mapping->getTargetBundle();
    $schema_type = $mapping->getSchemaType();

    $starterkits = [];
    $starterkits += ($this->starterKitTypes["$entity_type_id:$bundle:$schema_type"] ?? []);
    $starterkits += ($this->starterKitTypes["$entity_type_id:$schema_type"] ?? []);
    return $starterkits;
  }

  /**
   * Gets the relationships based on a Schema.org mapping.
   *
   * @param \Drupal\schemadotorg\SchemaDotOrgMappingInterface $mapping
   *   The Schema.org mapping.
   *
   * @return array
   *   An array of relationships.
   */
  protected function getRelationships(SchemaDotOrgMappingInterface $mapping): array {
    $relationships = [
      'hierarchy' => [],
      'relationships' => [],
      'links' => [],
      'enumerations' => [],
      'taxonomy_term' => [],
      'media' => [],
    ];

    $schema_properties = $mapping->getAllSchemaProperties();
    $field_definitions = $this->entityFieldManager
      ->getFieldDefinitions('node', $mapping->getTargetBundle());
    foreach ($field_definitions as $field_name => $field_definition) {
      $schema_property = $schema_properties[$field_name] ?? NULL;
      if (!$schema_property) {
        continue;
      }

      $relationship_type = $this->getRelationshipType($schema_property, $field_definition);
      if ($relationship_type) {
        $relationships[$relationship_type][$field_name] = $schema_property;
      }
    }

    return $relationships;
  }

  /**
   * Get the target Schema.org types or bundles for a given field definition.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   *
   * @return array
   *   An array of target Schema.org types or bundles.
   */
  protected function getTargets(FieldDefinitionInterface $field_definition): array {
    $settings = $field_definition->getSettings();
    $target_bundles = NestedArray::getValue($settings, ['handler_settings', 'target_bundles']);
    if (!$target_bundles) {
      return [];
    }

    /** @var \Drupal\schemadotorg\SchemaDotOrgMappingStorageInterface $mapping_storage */
    $mapping_storage = $this->entityTypeManager()
      ->getStorage('schemadotorg_mapping');
    $ids = $mapping_storage
      ->getQuery()
      ->condition('target_entity_type_id', 'node')
      ->condition('target_bundle', $target_bundles, 'IN')
      ->execute();
    /** @var \Drupal\schemadotorg\SchemaDotOrgMappingInterface[] $mappings */
    $mappings = $mapping_storage->loadMultiple($ids);

    foreach ($mappings as $mapping) {
      $target_bundles[$mapping->getTargetBundle()] = $mapping->getSchemaType();
    }
    return $target_bundles;
  }

  /**
   * Gets the relationship type for a given Schema.org property and field definition.
   *
   * @param string $schema_property
   *   The Schema.org property.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   *
   * @return string|null
   *   The relationship type, or NULL if none is found.
   */
  protected function getRelationshipType(string $schema_property, FieldDefinitionInterface $field_definition): ?string {
    $field_type = $field_definition->getType();
    if (str_contains($field_type, 'entity_reference')) {
      $target_type = $field_definition->getSetting('target_type');
      if (in_array($schema_property, $this->hierarchyProperties)) {
        return 'hierarchy';
      }
      elseif (in_array($target_type, ['taxonomy_term', 'media'])) {
        return $target_type;
      }
      else {
        return 'relationships';
      }
    }
    elseif (str_starts_with($field_type, 'list')) {
      return 'enumerations';
    }
    elseif ($field_type === 'link') {
      return 'links';
    }
    return NULL;
  }

  /**
   * Get Schema.org relationship types.
   *
   * @return array
   *   An array of Schema.org relationship types.
   */
  protected function getRelationshipTypes(): array {
    return [
      'hierarchy' => [
        'singular' => $this->t('Hierarchical'),
        'plural' => $this->t('Hierarchical'),
      ],
      'relationships' => [
        'singular' => $this->t('Relationship'),
        'plural' => $this->t('Relationships'),
      ],
      'links' => [
        'singular' => $this->t('Link'),
        'plural' => $this->t('Links'),
      ],
      'enumerations' => [
        'singular' => $this->t('Enumeration'),
        'plural' => $this->t('Enumerations'),
      ],
      'taxonomy_term' => [
        'singular' => $this->t('Taxonomy'),
        'plural' => $this->t('Taxonomies'),
      ],
      'media' => [
        'singular' => $this->t('Media'),
        'plural' => $this->t('Media'),
      ],
    ];
  }

}
